/*
 * Test Class for Heart
 */
package ch.zhaw.prog2.heartbeat;

import static org.junit.jupiter.api.Assertions.*;
import static org.mockito.Mockito.*;

import ch.zhaw.prog2.heartbeat.parts.Valve;
import org.junit.Before;
import org.junit.jupiter.api.BeforeAll;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Disabled;
import org.junit.jupiter.api.Test;

import ch.zhaw.prog2.heartbeat.Heart.State;
import org.junit.rules.ExpectedException;
import org.junit.runner.RunWith;
import org.mockito.InOrder;
import org.mockito.Mock;
import org.mockito.Mockito;
import org.mockito.MockitoAnnotations;
import org.mockito.junit.MockitoJUnitRunner;

@RunWith(MockitoJUnitRunner.class)
class HeartTest {
    /**
     * This is a very simple test to check if Junit and Mockito are properly set up.
     */
    @Test
    void testTheTest() {
        Heart classUnderTest = new Heart();
        assertNotNull(classUnderTest.getState(), "The heart must have a state.");
    }

    /**
     * Tests a single heartbeat
     */
    @Test
    void testHeartBeat() throws Heart.HeartBeatDysfunctionException, Valve.IllegalValveStateException, Heart.InvalidValvePositionException {
        Heart heart = new Heart();
        State startState = heart.getState();

        heart.executeHeartBeat();

        // after one heartbeat, the heart must be in the same state as before
        assertEquals(startState, heart.getState());
    }

    /**
     * Tests if the valves are open or closed depending on the status of the heart
     */
    @Test
    void testValveStatus() throws Heart.HeartBeatDysfunctionException, Valve.IllegalValveStateException, Heart.InvalidValvePositionException {
        Heart heart = new Heart();

        heart.executeHeartBeat();

        State state = heart.getState();

        if (state.equals(Heart.State.DIASTOLE)) {
            for (Half half : heart.getHalves()) {
                assertFalse(half.isAtrioventricularValveOpen());
                assertTrue(half.isSemilunarValveOpen());
            }
        } else if ((state.equals(Heart.State.SYSTOLE))) {
            for (Half half : heart.getHalves()) {
                assertTrue(half.isAtrioventricularValveOpen());
                assertFalse(half.isSemilunarValveOpen());
            }
        }
    }

    /**
     * Tests if the hart throws the appropriate Exception, when malfunction was detected during hartBeat
     */
    @Test
    void testExecuteHeartBeatErrorBehaviour() {
        Heart heart = new Heart();
        // prepare error situation due to wrong initialization
        heart.setState(State.SYSTOLE);

        assertThrows(Heart.HeartBeatDysfunctionException .class, // verification using lambda
            () -> heart.executeHeartBeat());
    }

    /**
     * Tests if the hart throws the appropriate Exception, when malfunction was detected during hartBeat
     * with exception Stubbing
     */
    @Test
    void testExecuteHartBeatErrorBehaviourWithStubbing() {
        //TODO implement
        Half half = mock(Half.class, withSettings().useConstructor(Half.Side.LEFT));
        Heart heart = new Heart(half, new Half(Half.Side.RIGHT));
        half.initializeState(heart.getState());
        when(half.isAtrioventricularValveOpen()).thenReturn(false);

        assertThrows(Heart.HeartBeatDysfunctionException .class, // verification using lambda
            () -> heart.executeHeartBeat());
    }

    /**
     * We test if Heart::executeHeartbeat() sends the right signals to both of its
     * halves.
     *
     * When Half::contractVentricle() is called, Half::closeAtrioventricularValve()
     * and Half::openSemilunarValve() must have been called earlier.
     *
     */
    @Test
    void testValvesBehavior() throws Valve.IllegalValveStateException, Heart.HeartBeatDysfunctionException, Heart.InvalidValvePositionException {
        //TODO implement
        Half half1 = mock(Half.class, withSettings().useConstructor(Half.Side.LEFT));
        Half half2 = mock(Half.class, withSettings().useConstructor(Half.Side.RIGHT));
        Heart heart = new Heart(half1, half2);

        heart.executeHeartBeat();

        InOrder inOrderHalf1 = inOrder(half1);

        inOrderHalf1.verify(half1).closeAtrioventricularValve();
        inOrderHalf1.verify(half1).openSemilunarValve();
        inOrderHalf1.verify(half1).relaxAtrium();
        inOrderHalf1.verify(half1).relaxVentricle();

        inOrderHalf1.verify(half1).closeAtrioventricularValve();
        inOrderHalf1.verify(half1).openSemilunarValve();
        inOrderHalf1.verify(half1).contractAtrium();
        inOrderHalf1.verify(half1).contractVentricle();

        inOrderHalf1.verifyNoMoreInteractions();

        InOrder inOrderHalf2 = inOrder(half2);

        inOrderHalf2.verify(half2).closeAtrioventricularValve();
        inOrderHalf2.verify(half2).openSemilunarValve();
        inOrderHalf2.verify(half2).relaxAtrium();
        inOrderHalf2.verify(half2).relaxVentricle();

        inOrderHalf2.verify(half2).closeAtrioventricularValve();
        inOrderHalf2.verify(half2).openSemilunarValve();
        inOrderHalf2.verify(half2).contractAtrium();
        inOrderHalf2.verify(half2).contractVentricle();

        inOrderHalf2.verifyNoMoreInteractions();

    }

    /**
     * This is code used for the lecture slide.
     */
    @Test
    void testForSlide() throws Valve.IllegalValveStateException, Heart.InvalidValvePositionException {
        Half mockedHalf = mock(Half.class);
        Heart heart = new Heart(mockedHalf, new Half(Half.Side.RIGHT));
        heart.setState(State.SYSTOLE);
        heart.initalizeState();

        when(mockedHalf.isAtrioventricularValveOpen()).thenReturn(false);
        when(mockedHalf.isSemilunarValveOpen()).thenReturn(true);
        heart.executeSystole();

        verify(mockedHalf).contractVentricle();
        verify(mockedHalf, times(1)).contractAtrium();
    }

}
